import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
from keras.datasets.fashion_mnist import load_data
from keras.models import Sequential
from keras.layers import Conv2D, Conv2DTranspose, MaxPool2D, GaussianNoise, Reshape
from keras.optimizers import SGD, RMSprop, Adam, Adadelta, Adamax, Adagrad

(X_train_full, y_train_full), (X_test, y_test) = load_data()
X_train_full = X_train_full.astype(np.float32) / 255
X_test = X_test.astype(np.float32) / 255
X_train, X_valid = X_train_full[:55000], X_train_full[55000:]
# y_train, y_valid = y_train_full[:55000], y_train_full[55000:]

# targets = ["T-shirt/top", "Trouser", "Pullover", "Dress", "Coat", "Sandal",
#            "Shirt", "Sneaker", "Bag", "Ankle boot"]
# n_rows = 3
# n_cols = 8
# plt.figure(figsize=(n_cols * 2, n_rows * 2))
# for row in range(n_rows):
#     for col in range(n_cols):
#         index = n_cols * row + col
#         plt.subplot(n_rows, n_cols, index + 1)
#         plt.imshow(X_train[index], cmap="binary")
#         plt.axis('off')
#         plt.title(targets[y_train[index]], fontsize=12)
# plt.subplots_adjust(wspace=0.2, hspace=0.5)
# plt.show()

denoising_encoder = Sequential([
    Reshape([28, 28, 1], input_shape=[28, 28]),
    GaussianNoise(0.2),
    Conv2D(16, kernel_size=3, padding="SAME", activation="selu"),
    MaxPool2D(pool_size=2),
    Conv2D(32, kernel_size=3, padding="SAME", activation="selu"),
    MaxPool2D(pool_size=2),
    Conv2D(64, kernel_size=3, padding="SAME", activation="selu"),
    MaxPool2D(pool_size=2)
])

# tf.keras.utils.plot_model(denoising_encoder, to_file='denoising_encoder.png', show_shapes=True)

denoising_decoder = Sequential([
    Conv2DTranspose(32, kernel_size=3, strides=2, padding="VALID", activation="selu",
                    input_shape=[3, 3, 64]),
    Conv2DTranspose(16, kernel_size=3, strides=2, padding="SAME", activation="selu"),
    Conv2DTranspose(1, kernel_size=3, strides=2, padding="SAME", activation="sigmoid"),
    Reshape([28, 28])
])

# tf.keras.utils.plot_model(denoising_decoder, to_file='denoising_decoder.png', show_shapes=True)

denoising_ae = Sequential([denoising_encoder, denoising_decoder])
denoising_ae.compile(loss="binary_crossentropy", optimizer=SGD(lr=0.5))
history = denoising_ae.fit(X_train, X_train, epochs=5,
                           validation_data=(X_valid, X_valid))

corrupted_inputs = GaussianNoise(0.2)(X_valid[8:13], training=True)
reconstructs = denoising_ae.predict(corrupted_inputs)
plt.figure(figsize=(15, 12))
for i in range(5):
    plt.subplot(2, 5, 1 + i)
    plt.imshow(corrupted_inputs[i], cmap='binary')
    plt.axis('off')
    plt.subplot(2, 5, 6 + i)
    plt.imshow(reconstructs[i], cmap='binary')
    plt.axis('off')
plt.show()
